/************************************************************************************/
/*  Copyright (c) 2017 by SAS Institute, Inc., Cary, NC, USA.                       */
/*  All rights reserved.                                                            */
/************************************************************************************/
/*  NAME:        rskmgt_add_cust_fields.sas                                         */
/*  VERSION:     7.3                                                                */
/*  DESCRIPTION: Add custom fields to table.                                        */
/*  PARAMETERS: indsn=target table                                                  */
/*              feature=business_object_nm in cust_field_value_l                    */
/*              rk_column=business_object_rk in cust_field_value_l                  */
/*              business_object_type_nm =business_object_type_nm in cust_field_defn */
/*              filter_lst=pulls only the custom fields listed in this parameter    */
/*                         if blank, all custom fields are pulled; otherwise the    */
/*                         list should be comma delimited                           */ 
/*              MRA=Y or N (default) - pulls the max cust_obj_14_rk custom field    */
/*                  value for customObject(14) (aka MRA) to only store the latest   */
/*                  ratings.  This is currently only applicable when adding these   */
/*                  custom fields to the customObject1 (aka Model) object.          */
/*  PRODUCT:    SAS Risk Management Server                                          */
/************************************************************************************/

%macro rskmgt_add_cust_fields(indsn=, 
                              feature=,
                              rk_column=,
                              business_object_type_nm=,
                              filter_lst=,
                              MRA=N);

%put NOTE: {START: &sysmacroname};

/* Get the custom object rest call names */
%rskmgt_get_rest_obj_names(business_obj_type_nm=&business_object_type_nm);

/* Construct the URL for the rest call */
%let fullurl=&srv_url.&&&prod.service_name./rest/&restobj.?start=0%str(&)limit=&maxLimit%str(&)fields=links,customFields;

/* Call the rest service to pull the business object custom fields and links */
filename svc_rsp temp;
filename headout temp;
filename jmap temp;

proc http
  url= "&fullurl."
  method="GET"
  out=svc_rsp
  headerout=headout
  HTTP_TOKENAUTH ;
run;

/* JSON libname to the SAS datasets produced from the rest call */
libname posts JSON fileref=svc_rsp map=jmap automap=create;

/* Copy all JSON response datasets to work */
proc copy in=posts out=work;
  select items root;
run;

/* Get the total # of records for this custom object */
proc sql noprint;
  select count into : custField_cnt
  from work.root;
quit;

/* Initialize the Do Loop counter */
%let iLimit=;

/* Get a list of all custom fields that have a value */
%let cust_def_cnt=;

proc sql noprint;
  create table custom_field_defn as
  select distinct a.cust_field_nm, cust_field_type_cd, cust_field_max_length, cust_field_desc
  from &src_lib..cust_field_defn_l a join
       &src_lib..cust_field_value_l b
  on (upcase(a.cust_field_nm)=upcase(b.cust_field_nm) and
      upcase(a.business_object_type_nm)="%upcase(&business_object_type_nm)" and
      upcase(b.business_object_nm)="%upcase(&feature)" and
      (value_dt ne . OR value_no ne . OR value_cd ne ' ' OR value_txt ne ' ' OR
       value_flg ne ' '))
  order by a.cust_field_nm;

  select count(*) into : cust_def_cnt
  from custom_field_defn;
quit;

/* If there are custom fields, proceed.  Otherwise, exit the macro. */
%if &cust_def_cnt = 0 %then
%do;
   %goto no_custom_flds;
%end;
   
/* Add labels to custom fields */
data work.custom_field_defn;
  length messageid $200;
  set work.custom_field_defn;
  messageid="&business_object_type_nm"||'_'||'OBJ'||'_'||trim(left(cust_field_nm));
run;

/* In case there are duplicates, remove any duplicates from custommessages before doing the update below */
proc sort tagsort data=custommessages nodupkey;
  by key;
run;

/* Update the custom_field_desc w/ the labels from custommessages */
proc sql noprint;
  update custom_field_defn
  set cust_field_desc=
    (select unicode(text,'esc')
     from custommessages
     where upcase(Custommessages.key)=upcase(Custom_field_defn.messageid) and
          (upcase(Custommessages.locale)="%upcase(&localecd)" or
           upcase(Custommessages.locale)="%substr(%upcase(&localecd),1,2)"));
quit;

proc sort tagsort data=custom_field_defn out=custom_field_defn(drop=messageid);
   by cust_field_nm;
run;

/* Create a list of all possible DTE,OPM,OPS,BOL,NUM & STR custom field names, lengths and labels */
%let dte_custfld_lst=;
%let num_custfld_lst=;
%let opm_custfld_lst=;
%let ops_custfld_lst=;
%let bol_custfld_lst=;
%let str_custfld_lst=;

proc sql noprint;
  select distinct quote(trim(cust_field_nm)), quote(trim(cust_field_desc)) into : dte_custfld_lst separated by ',',
                                                                                 : dte_label_lst separated by '~'
  from custom_field_defn
  where cust_field_type_cd='DTE';

  select distinct quote(trim(cust_field_nm)), quote(trim(cust_field_desc)) into : num_custfld_lst separated by ',',
                                                                                 : num_label_lst separated by '~'
  from custom_field_defn
  where cust_field_type_cd='NUM';

  select distinct quote(trim(cust_field_nm)), cust_field_max_length, quote(trim(cust_field_desc)) into : opm_custfld_lst separated by ',',
                                                                                                        : opm_max_len_lst separated by ',',
                                                                                                        : opm_label_lst separated by '~'
  from custom_field_defn
  where cust_field_type_cd='OPM';

  select distinct quote(trim(cust_field_nm)), cust_field_max_length, quote(trim(cust_field_desc)) into : ops_custfld_lst separated by ',',
                                                                                                        : ops_max_len_lst separated by ',',
                                                                                                        : ops_label_lst separated by '~'
  from custom_field_defn
  where cust_field_type_cd='OPS';

  select distinct quote(trim(cust_field_nm)), quote(trim(cust_field_desc)) into : bol_custfld_lst separated by ',',
                                                                                 : bol_label_lst separated by '~'
  from custom_field_defn
  where cust_field_type_cd='BOL';

  select distinct quote(trim(cust_field_nm)), cust_field_max_length, quote(trim(cust_field_desc)) into : str_custfld_lst separated by ',',
                                                                                                        : str_max_len_lst separated by ',',
                                                                                                        : str_label_lst separated by '~'
  from custom_field_defn
  where cust_field_type_cd='STR';
quit;

%let DBCSVal=%sysfunc(getoption(DBCS));

/* Loop to pull all custom fields by business object */ 
%do %until (&iLimit >= &custField_cnt);
   
  /* Copy all JSON response datasets to work */
  proc copy in=posts out=work;
  run;
   
  /* Check to see if there are any records in the links table */
  /* If so, continue; otherwise, exit the macro */
  %rskmgt_ds_exists(indsn=work.items_links,
                    output_nobs=items_links_cnt);
                    
  %if &items_links_cnt <= 0 %then %goto customFields;
    
  /* Flip the OPS/OPM fields into columns */
  proc transpose data=work.items_links out=work.items_links_trans(drop=_NAME_) let;
    var title;
    id rel;
    by ordinal_items;
  run;
  quit;
    
  /* Change all OPM pipe delimited fields to comma delimited */
  data work.items_links_fnl;
    set work.items_links_trans;
    %if &opm_custfld_lst ne %str() %then
    %do;
      %let i=1;
      %let opm_field_nm=%scan(%bquote(&opm_custfld_lst),&i,%str(","));
      %do %while (&opm_field_nm ne );
        %local dsid rc;
        %let dsid=%sysfunc(open(items_links_trans));
        %if %sysfunc(varnum(&dsid,&opm_field_nm)) > 0 %then
        %do;
          &opm_field_nm=tranwrd(&opm_field_nm,'|',',');
        %end;
        %let rc=%sysfunc(close(&dsid));

        %let i=%eval(&i+1);
        %let opm_field_nm=%scan(%bquote(&opm_custfld_lst),&i,%str(","));
      %end;
    %end;
    drop creator updater sourceSystemCd self delete update dimensionalPoints;
  run;
  
  %customFields:
  
  /* Check to see if there are any records in the customFields table */
  /* If so, continue; otherwise, exit the macro */
  %rskmgt_ds_exists(indsn=work.items_customfields,
                    output_nobs=items_customfields_cnt);
                    
  %if &items_customfields_cnt <= 0 %then %goto noCustomFieldVals;
   
  /* Get a list of all custom fields in the REST call */
  proc contents data=posts.items_customfields memtype=data
                out=work.items_customfields_trans(rename=(name=cust_field_nm)) noprint;
  run;

  /* Filter down the list DTE,NUM,BOL and OPS fields to what is availabe in this rest call */
  %let date_lst=;
  %let num_lst=;
  %let ops_lst=;
  %let bol_lst=;

  proc sql noprint;
     
  %if &dte_custfld_lst ne %str() %then
  %do;
    select distinct cust_field_nm into : date_lst separated by ' '
    from work.items_customfields_trans
    where substr(upcase(cust_field_nm),1,7) ne 'ORDINAL' and
          cust_field_nm in %upcase((&dte_custfld_lst));
  %end;
     
  %if &num_custfld_lst ne %str() %then
  %do;
    select distinct cust_field_nm into : num_lst separated by ' '
    from work.items_customfields_trans
    where substr(upcase(cust_field_nm),1,7) ne 'ORDINAL' and
          cust_field_nm in %upcase((&num_custfld_lst));

  %end;
     
  %if &bol_custfld_lst ne %str() %then
  %do;
    select distinct cust_field_nm into : bol_lst separated by ' '
    from work.items_customfields_trans
    where substr(upcase(cust_field_nm),1,7) ne 'ORDINAL' and
          cust_field_nm in %upcase((&bol_custfld_lst));

  %end;
     
  %if &ops_custfld_lst ne %str() %then
  %do;
    select distinct cust_field_nm into : ops_lst separated by ' '
    from work.items_customfields_trans
    where substr(upcase(cust_field_nm),1,7) ne 'ORDINAL' and
          cust_field_nm in %upcase((&ops_custfld_lst));
   %end;
     
   quit;

   /* Cast the DTE and NUM fields from character to numeric fields w/ formats and informats */
   /* Cast the BOL field from numeric to character */
   /* Rename the OPS coded fields */     
   data work.items_customFields;
     set posts.items_customfields;

     %if &date_lst ne %str() %then
     %do;
       %let h=1;
       %let dte_field_nm=%scan(%bquote(&date_lst),&h,' ');
       %do %while (&dte_field_nm ne);
         d_&dte_field_nm=input(&dte_field_nm,yymmdd10.);
         informat d_&dte_field_nm nldate.;
         format d_&dte_field_nm nldate.;
         drop &dte_field_nm;
         rename d_&dte_field_nm=&dte_field_nm;

          %let h=%eval(&h+1);
          %let dte_field_nm=%scan(%bquote(&date_lst),&h,' ');
       %end;
     %end;

     %if &num_lst ne %str() %then
     %do;
       %let i=1;
       %let num_field_nm=%scan(%bquote(&num_lst),&i,' ');
       %do %while (&num_field_nm ne);
         n_&num_field_nm=input(&num_field_nm,nlnum19.);
         informat n_&num_field_nm nlnum19.;
         format n_&num_field_nm nlnum19.;
         drop &num_field_nm;
         rename n_&num_field_nm=&num_field_nm;

         %let i=%eval(&i+1);
         %let num_field_nm=%scan(%bquote(&num_lst),&i,' ');
       %end;
     %end;

     %if &bol_lst ne %str() %then
     %do;
       %let l=1;
       %let bol_field_nm=%scan(%bquote(&bol_lst),&l,' ');
       %do %while (&bol_field_nm ne);
         b_&bol_field_nm=put(&bol_field_nm,1.);
         if b_&bol_field_nm='0' then b_&bol_field_nm='N';
         if b_&bol_field_nm='1' then b_&bol_field_nm='Y';
         drop &bol_field_nm;
         rename b_&bol_field_nm=&bol_field_nm;

         %let l=%eval(&l+1);
         %let bol_field_nm=%scan(%bquote(&bol_lst),&l,' ');
       %end;
     %end;

     %if &ops_lst ne %str() %then
     %do;
       %let j=1;
       %let ops_field_nm=%scan(%bquote(&ops_lst),&j,' ');
       %do %while (&ops_field_nm ne);
         length &ops_field_nm._c

         %if &DBCSVal=DBCS %then
         %do;
           $%eval(32 * &char_multiplier);
         %end;
         %else
         %do;
           $%eval(32 * &char_multiplier);
         %end;
         &ops_field_nm._c=&ops_field_nm;
         drop &ops_field_nm;

         %let j=%eval(&j+1);
         %let ops_field_nm=%scan(%bquote(&ops_lst),&j,' ');
       %end;
     %end;
   run;

   /* Look for OPM generated datasets in the posts library */
   /* and create the OPM coded variables */
   %if &opm_custfld_lst ne %str() %then
   %do;
     %let o=1;
     %let opm_field_nm=%scan(%bquote(&opm_custfld_lst),&o,%str(","));

     %do %while (&opm_field_nm ne );

       /* REST call dataset name */
       %if %length(customFields_&opm_field_nm) >= 30 %then
       %do;
         %let in_custfld_dsn=%substr(customfields_&opm_field_nm,1,30);
         %let out_custfld_dsn=&opm_field_nm;
       %end;
       %else
       %do;
         %let in_custfld_dsn=customfields_&opm_field_nm;
         %let out_custfld_dsn=&opm_field_nm;
       %end;

       %if %sysfunc(exist(posts.&in_custfld_dsn)) %then
       %do;
         /* Get a list of all OPM columns in the dataset */
         proc contents data=posts.&in_custfld_dsn memtype=data
                       out=work.&out_custfld_dsn noprint;
         run;

         proc sql noprint;
           select distinct name, count(*), sum(length) into : opm_field_array_nms separated by ' ',
                                                            : opm_field_array_cnt,
                                                            : opm_val_length
           from work.&out_custfld_dsn
           where upcase(name) not like 'ORDINAL_%';
         quit;

         /* Create the OPM coded fields */
         data work.&out_custfld_dsn;
           length &opm_field_nm._c $%eval((&opm_val_length * &opm_field_array_cnt * &char_multiplier) + (&opm_field_array_cnt - 1));

           set posts.&in_custfld_dsn;

           %let b=1;
           %let cust_field_array_nm=%scan(%bquote(&opm_field_array_nms),&b,' ');          
       
           %do %while (&cust_field_array_nm ne );
             %if &b=1 %then 
             %do;
                &opm_field_nm._c=strip(&cust_field_array_nm);
             %end;
             %else
             %do;
               if strip(&cust_field_array_nm) ne ' ' then
               do;
                 &opm_field_nm._c=strip(&opm_field_nm._c)||","|| strip(&cust_field_array_nm);
               end;
             %end;
                 
             %let b=%eval(&b+1);
             %let cust_field_array_nm=%scan(%bquote(&opm_field_array_nms),&b,' ');
          
           %end;
         run;

         data work.items_customFields;
           merge work.items_customFields work.&out_custfld_dsn.(keep=ordinal_customFields &opm_field_nm._c);
           by ordinal_customFields;
         run;
       %end;

       %let o=%eval(&o+1);
       %let opm_field_nm=%scan(%bquote(&opm_custfld_lst),&o,%str(","));
      %end;
    %end;

    %noCustomFieldVals:
    
    /* If there are links or customFields, merge together */
    %if %sysfunc(exist(work.items_customFields)) or %sysfunc(exist(work.items_links_fnl)) %then
    %do;
      /* Merge all of the businessObjects and customFields into one table */
      data work.bo_customFields(rename=(key=&rk_column) drop=ordinal_root ordinal_items version ordinal_customFields);
        merge work.items  
        
          %if &items_customfields_cnt > 0 %then
          %do;
            work.items_customFields
          %end;
        
          %if &items_links_cnt > 0 %then
          %do;
            work.items_links_fnl
          %end;
        %str(;)  
        by ordinal_items;
      run;
    %end;
     
    /* If this is the first iteration through the loop, create the output table structure */
    %if &iLimit eq %str() %then
    %do;
      /* Create the master table structure */
      data custfld_master;
        attrib &rk_column length=8.
        
        /* Create the STR, OPM and OPS columns */
        %let field_types=str,opm,ops;

        %do j=1 %to 3;

          %let field_type_nm=%scan(%bquote(&field_types),&j,',');

          %if &&&field_type_nm._custfld_lst ne %str() %then
          %do;
            %let h=1;
            %let &&field_type_nm._field_lst_nm=%scan(%bquote(&&&field_type_nm._custfld_lst),&h,%str(","));
            %let &&field_type_nm._field_lst_len=%scan(%bquote(&&&field_type_nm._max_len_lst),&h,',');
            %let &&field_type_nm._label_lst_desc=%qscan(%bquote(&&&field_type_nm._label_lst),&h,"~");

            %do %while (&&&field_type_nm._field_lst_nm ne );
              &&&field_type_nm._field_lst_nm

              %if &&&field_type_nm._field_lst_len eq . %then
              %do;
                %if &DBCSVal=DBCS %then
                %do;
                  length=$%eval(1000 * &char_multiplier) label="&&&field_type_nm._label_lst_desc"
                %end;
                %else
                %do;
                  length=$%eval(4000 * &char_multiplier) label="&&&field_type_nm._label_lst_desc"
                %end;
              %end;
              %else
              %do;
                length=$%eval(&&&field_type_nm._field_lst_len * &char_multiplier) label="&&&field_type_nm._label_lst_desc"
              %end;

              %if &&field_type_nm ne str %then
              %do;
                &&&field_type_nm._field_lst_nm._c length=$%eval(32 * &char_multiplier) label="&&&field_type_nm._label_lst_desc %sysfunc(sasmsg(sashelp.rskmgtsvr,Code, NOQUOTE))"
               %end;

              %let h=%eval(&h+1);
              %let &&field_type_nm._field_lst_nm=%scan(%bquote(&&&field_type_nm._custfld_lst),&h,%str(","));
              %let &&field_type_nm._field_lst_len=%scan(%bquote(&&&field_type_nm._max_len_lst),&h,',');
              %let &&field_type_nm._label_lst_desc=%qscan(%bquote(&&&field_type_nm._label_lst),&h,"~");
            %end;
          %end;
        %end;

        /* Create the DTE columns */
        %if &dte_custfld_lst ne %str() %then
        %do;
          %let k=1;
          %let dte_field_list_nm=%scan(%bquote(&dte_custfld_lst),&k,%str(","));
          %let dte_label_lst_desc=%qscan(%bquote(&dte_label_lst),&k,"~");

          %do %while (&dte_field_list_nm ne );
            &dte_field_list_nm length=8. format=NLDATE. informat=NLDATE. label="&dte_label_lst_desc"

            %let k=%eval(&k+1);
            %let dte_field_list_nm=%scan(%bquote(&dte_custfld_lst),&k,%str(","));
            %let dte_label_lst_desc=%qscan(%bquote(&dte_label_lst),&k,"~");
          %end;
        %end;

        /* Create the NUM columns */
        %if &num_custfld_lst ne %str() %then
        %do;
          %let l=1;
          %let num_field_list_nm=%scan(%bquote(&num_custfld_lst),&l,%str(","));
          %let num_label_lst_desc=%qscan(%bquote(&num_label_lst),&l,"~");

          %do %while (&num_field_list_nm ne );
            &num_field_list_nm length=8. format=NLNUM19.2 informat=NLNUM19.2 label="&num_label_lst_desc"

            %let l=%eval(&l+1);
            %let num_field_list_nm=%scan(%bquote(&num_custfld_lst),&l,%str(","));
            %let num_label_lst_desc=%qscan(%bquote(&num_label_lst),&l,"~");
          %end;
        %end;

        /* Create the BOL columns */
        %if &bol_custfld_lst ne %str() %then
        %do;
          %let m=1;
          %let bol_field_list_nm=%scan(%bquote(&bol_custfld_lst),&m,%str(","));
          %let bol_label_lst_desc=%qscan(%bquote(&bol_label_lst),&m,"~");

          %do %while (&bol_field_list_nm ne );
            &bol_field_list_nm length=$%eval(1 * &char_multiplier) label="&bol_label_lst_desc"

            %let m=%eval(&m+1);
            %let bol_field_list_nm=%scan(%bquote(&bol_custfld_lst),&m,%str(","));
            %let bol_label_lst_desc=%qscan(%bquote(&bol_label_lst),&m,"~");
          %end;
        %end;
        %str(;)
        stop;
      run;

      /* If work.bo_customFields, merge w/ custfld_master */
      %if %sysfunc(exist(work.bo_customFields)) %then
      %do;
        data work.custfld_master;
          set work.custfld_master work.bo_customFields;
        run;
      %end;
    %end;
    %else
    %do;
      /* If work.bo_customFields, merge w/ custfld_master */
      %if %sysfunc(exist(work.bo_customFields)) %then
      %do;
        data work.custfld_master;
          set work.custfld_master work.bo_customFields;
        run;
      %end;
    %end;

    %let iLimit=%eval(&iLimit+&maxLimit);

    %let fullurl=&srv_url.&&&prod.service_name./rest/&restobj.?start=&iLimit%str(&)limit=&maxLimit%str(&)fields=links,customFields;

    libname posts clear;

    proc http
      url="&fullurl."
      method="GET"
      out=svc_rsp
      headerout=headout
      HTTP_TOKENAUTH;
    run;

    libname posts JSON fileref=svc_rsp map=jmap automap=create;
    
  %end; /* %do %until (&iLimit >= &custField_cnt) loop */

  /* Add a custom field for status_dttm if X_STATUSCD exist */
  %let x_statuscd_exist=;

  proc sql noprint;
    select cust_field_nm into : x_statuscd_exist
    from work.custom_field_defn
    where upcase(cust_field_nm)='X_STATUSCD';
  quit;

  %if &x_statuscd_exist ne %str() %then
  %do;
    proc sql;
      create table status_dttm_vals as
      select &rk_column, b.valid_from_dttm as status_dttm
      from work.custfld_master a left join
           &src_lib..cust_field_value_l b
      on a.&rk_column=b.business_object_rk and
         b.business_object_nm="CUST_OBJ_1" and
         upcase(b.cust_field_nm)='X_STATUSCD'
      order by a.&rk_column;
    quit;

    data work.custfld_master;
      merge work.custfld_master work.status_dttm_vals;
      by &rk_column;
    run;
  %end;

  /* Update the MRA custom field values to include the latest MRA values for the Models data source only */
  %if &MRA=Y %then
  %do;
    %let business_object_type_nm=customObject14;

    proc sql noprint;
      create table custom_field_defn as
      select distinct cust_field_nm, cust_field_type_cd, cust_field_max_length, cust_field_desc
      from &src_lib..cust_field_defn_l
      where upcase(business_object_type_nm)="CUSTOMOBJECT14"

      %if &filter_lst ne %str() %then
      %do;
        %let h=1;
        %let filter_list_nm=%scan(&filter_lst,&h,',');
        %do %while (&filter_list_nm ne );
          %if &h=1 %then
          %do;
            AND (cust_field_nm="&filter_list_nm"
          %end;
          %else
          %do;
            OR cust_field_nm="&filter_list_nm"
          %end;

          %let h=%eval(&h+1);
          %let filter_list_nm=%scan(&filter_lst,&h,',');
        %end;
        )
      %end;

      order by cust_field_nm;
    quit;

    /* Add labels to the custom fields */
    data work.custom_field_defn;
      length messageid $200;
      set work.custom_field_defn;
      messageid="&business_object_type_nm"||'_'||'OBJ'||'_'||trim(left(cust_field_nm));
    run;

    /* In case there are duplicates, remove any duplicates from custommessages before doing the update below */
    proc sort tagsort data=Custommessages nodupkey;
      by key;
    run;

    /* Update the custom_field_desc w/ the labels from custommessages */
    proc sql noprint;
      update custom_field_defn
      set cust_field_desc=
        (select unicode(text,'esc')
         from custommessages
         where upcase(Custommessages.key)=upcase(Custom_field_defn.messageid) and
              (upcase(Custommessages.locale)="%upcase(&localecd)" or
               upcase(Custommessages.locale)="%substr(%upcase(&localecd),1,2)"));
    quit;

    proc sort tagsort data=custom_field_defn out=cust_field_list(drop=messageid) nodupkey;
      by cust_field_nm;
    run;

    /* Merge in the STR, BOL, DTE, OPS and NUM MRA coded value fields w/ the models */
    filename mratmpa temp;

    data _null_;
      length str_len ops_len bol_len 8. cust_field_cd $32;
      set work.cust_field_list;

      file mratmpa;

      put 'proc sql noprint;';
      put 'create table work.custfld_master as';
      put 'select A.*,';

      if cust_field_type_cd='STR' then
      do;
        put 'cust_field_value_l.value_txt as';
        put cust_field_nm;

        if cust_field_max_length ne . then
        do;
          put 'length=';
          str_len=cust_field_max_length * &char_multiplier;
          put str_len;
          put "label='";
          put cust_field_desc;
          put "'";
        end;
        else
        do;
          %if &DBCSVal=DBCS %then
          %do;
            put 'length=';
            str_len=1000 * &char_multiplier;
            put str_len;
            put "label='";
            put cust_field_desc;
            put "'";
          %end;
          %else
          %do;
            put 'length=';
            str_len=4000 * &char_multiplier;
            put str_len;
            put "label='";
            put cust_field_desc;
            put "'";
          %end;
        end;
      end;

      if cust_field_type_cd='OPS' then
      do;
        put 'cust_field_value_l.value_cd as';
        cust_field_cd=strip(cust_field_nm)||'_c';
        put cust_field_cd;

        put 'length=';
        ops_len=32 * &char_multiplier;
        put ops_len;
        put "label='";
        put cust_field_desc;
        put ' ';
        put "%sysfunc(sasmsg(sashelp.rskmgtsvr,Code, NOQUOTE))";
        put "'";
      end;

      if cust_field_type_cd='BOL' then
      do;
        put 'cust_field_value_l.value_flg as';
        put cust_field_nm;
        put 'length=';
        bol_len=1 * &char_multiplier;
        put bol_len;
        put "label='";
        put cust_field_desc;
      end;

      if cust_field_type_cd='NUM' then
      do;
        put 'cust_field_value_l.value_num as';
        put cust_field_nm;
        put 'length=8 format=NLNUM19.2 informat=NLNUM19.2';
        put "label='";
        put cust_field_desc;
      end;

      if cust_field_type_cd='DTE' then
      do;
        put 'cust_field_value_l.value_dt as';
        put cust_field_nm;
        put 'length=8 format=NLDATE. informat=NLDATE.';
        put "label='";
        put cust_field_desc;
      end;

      put 'from work.custfld_master A left join';
      put '(select B.cust_obj_1_rk,';
      put '        max(business_object_rk_1) as cust_obj_14_rk';
      put " from &src_lib..cust_obj_1_l B,";
      put "      &src_lib..link_instance_l L,";
      put "      &src_lib..link_type_l T,";
      put "      &src_lib..cust_field_value_l X";
      put ' where';
      put '   T.link_type_rk=L.link_type_rk and';
      put "   T.link_type_id='modelRiskAssmt_model' and ";
      put '   L.business_object_rk_2=B.cust_obj_1_rk';
      put ' group by';
      put '   B.cust_obj_1_rk';
      put ' ) LAST_MRA';
      put " on A.&rk_column=LAST_MRA.cust_obj_1_rk";
      put " left join &src_lib..cust_field_value_l";
      put ' on LAST_MRA.cust_obj_14_rk=cust_field_value_l.business_object_rk and';
      put '    cust_field_value_l.business_object_nm=';
      put "'";
      put 'CUST_OBJ_14';
      put "'";   
      put "and cust_field_value_l.cust_field_nm='";
      put cust_field_nm;
      put "';";
      put 'quit;';
    run;

    %include mratmpa;
    filename mratmpa clear;

    /* Create a list of all MRA cust fields w/ OPS label values */
    proc sql;
      create table mra_ops_labels as
      select distinct c.cust_field_nm,
             c.cust_field_type_cd,
             c.cust_field_max_length,
             n.list_nm,
             n.option_cd,
             f.desc_txt
      from &src_lib..cust_field_defn_l c left join
           &src_lib..named_list_mapping_l m
      on upcase(c.business_object_type_nm)="%upcase(&business_object_type_nm)"
      and c.cust_field_nm=m.field_nm
      and upcase(c.cust_field_type_cd)='OPS'

      %if &filter_lst ne %str() %then
      %do;
        %let h=1;
        %let filter_list_nm=%scan(&filter_lst,&h,',');
        %do %while (&filter_list_nm ne );
          %if &h=1 %then
          %do;
            AND (c.cust_field_nm="&filter_list_nm"
          %end;
          %else
          %do;
            OR c.cust_field_nm="&filter_list_nm"
          %end;
          %let h=%eval(&h+1);
          %let filter_list_nm=%scan(&filter_lst,&h,',');
        %end;
        )
      %end;

      left join
      &src_lib..named_list_l n
      on m.list_nm=n.list_nm
      join
      &src_lib..full_bo_trans f
      on n.named_list_rk=f.business_object_rk
      and upcase(f.business_object_nm)='NAMED_LIST'
      and (upper(f.language_cd)="%upcase(&localecd)"
      or upcase(f.language_cd)="%substr(%upcase(&localecd),1,2)")
      order by c.cust_field_nm, n.list_nm, n.option_cd;
    quit;

    filename mratmpb temp;

    /* Merge in the OPS MRA label value fields w/ the models */
    data _null_;
      length ops_len 8. cust_field_cd $32;
      set work.cust_field_list;
      where cust_field_type_cd='OPS';

      cust_field_cd=strip(cust_field_nm)||'_c';

      file mratmpb;

      put 'proc sql noprint;';
      put 'create table work.custfld_master as';
      put 'select A.*,';

      put 'mra.desc_txt as';
      put cust_field_nm;

      if cust_field_max_length ne . then
      do;
        ops_len=cust_field_max_length * &char_multiplier;
        put 'length=';
        put ops_len;
        put "label='";
        put cust_field_desc;
        put "'";
      end;
      else
      do;
        %if &DBCSVal=DBCS %then
        %do;
          put 'length=';
          ops_len=1000 * &char_multiplier;
          put ops_len;
          put "label='";
          put cust_field_desc;
          put "'";
        %end;
        %else
        %do;
          put 'length=';
          ops_len=4000 * &char_multiplier;
          put ops_len;
          put "label='";
          put cust_field_desc;
          put ' ';
          put "%sysfunc(sasmsg(sashelp.rskmgtsvr,Code, NOQUOTE))";
          put "'";
        %end;
      end;

      put 'from work.custfld_master a left join';
      put 'work.mra_ops_labels mra';
      put 'on';
      put  cust_field_cd;
      put '=';
      put 'mra.option_cd and';
      put 'mra.cust_field_nm';
      put '=';
      put "'";
      put cust_field_nm;
      put "';";
      put 'quit;';
    run;

    %include mratmpb;
    filename mratmpb clear;
   %end;
   
   /* Merge the customFields w/ the &indsn data */
   /* Perform sort 1st */
   proc sort tagsort data=work.custfld_master;
     by &rk_column;
   run;
   
   proc sort tagsort data=&indsn;
     by &rk_column;
   run;
   
   data &indsn;
     merge &indsn work.custfld_master;
     by &rk_column;
    run;

   %no_custom_flds:

%put NOTE: {END: &sysmacroname};

%mend rskmgt_add_cust_fields;